﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using UnityEngine;
using TriangleNet;
using TriangleNet.Geometry;
using TriangleNet.Topology;
using UnityEngine.Profiling;

public class DataCollection : MonoBehaviour
{
	int testIndex = -1;

	int NUMBER_OF_RAYS = 3200;
    float SPEED = 10f;
    float DISTANCE_TRESH = 0.7f;
    int ANGLE_THRESH = 40;
    Algorithm algorithm;
    string LOG_PATH = "";

	List<TestInstance> tests;

	bool debug = true;

	Ray playerLookingRay;
	List<Vector3> grid = new List<Vector3> ();
	List<Vector3> backgroundGrid = new List<Vector3> ();
	Vector2 windowSize;
	TriangleNet.Mesh mesh;
	float rotation;
	int secondstofreeze = 5;
	float secondsFrozen = 0;
	float deltaTime = 0.0f;


	public Dictionary<int, ObjectData> visibleObjects = new Dictionary<int, ObjectData> ();
	public Dictionary<int, ObjectData> backgroundObjects = new Dictionary<int, ObjectData> ();

	public List<JSONEntry> buffer = new List<JSONEntry> ();

	void Awake ()
	{
		QualitySettings.vSyncCount = 0;  // VSync must be disabled
		Application.targetFrameRate = 300;
	}

void Start()
    {
        tests = new List<TestInstance>()
        {
            new TestInstance(Algorithm.GOLD,2000)
            /*
            new TestInstance(Algorithm.Baseline, 800),
            new TestInstance(Algorithm.Baseline, 900),
            new TestInstance(Algorithm.Baseline, 1000),
            new TestInstance(Algorithm.Baseline, 1100),
            new TestInstance(Algorithm.Baseline, 1200),
            new TestInstance(Algorithm.Baseline, 1300),
            new TestInstance(Algorithm.Baseline, 1400),
            new TestInstance(Algorithm.Baseline, 1500),
            new TestInstance(Algorithm.Baseline, 1600),
            new TestInstance(Algorithm.Baseline, 1700),
            new TestInstance(Algorithm.Baseline, 1800),
            new TestInstance(Algorithm.Baseline, 1900),
            new TestInstance(Algorithm.Baseline, 2000),
            new TestInstance(Algorithm.Baseline, 2100),
            new TestInstance(Algorithm.Baseline, 2200),
            new TestInstance(Algorithm.Baseline, 2300),
            new TestInstance(Algorithm.Baseline, 2400),
            new TestInstance(Algorithm.Baseline, 2500),
            new TestInstance(Algorithm.Baseline, 2600),
            new TestInstance(Algorithm.Baseline, 2700),
            new TestInstance(Algorithm.Baseline, 2800),
            new TestInstance(Algorithm.Baseline, 2900),
            new TestInstance(Algorithm.Baseline, 3000),
            new TestInstance(Algorithm.Baseline, 3100),
            new TestInstance(Algorithm.Baseline, 3200),
            */

            /*
            new TestInstance(Algorithm.A, 800, 0.5f),
            new TestInstance(Algorithm.A, 800, 0.7f),
            new TestInstance(Algorithm.A, 800, 0.9f),
            
            new TestInstance(Algorithm.A, 1600, 0.5f),
            new TestInstance(Algorithm.A, 1600, 0.7f),
            new TestInstance(Algorithm.A, 1600, 0.9f),
            
            new TestInstance(Algorithm.A, 2400, 0.5f),
            new TestInstance(Algorithm.A, 2400, 0.7f),
            new TestInstance(Algorithm.A, 2400, 0.9f),
            */

            //new TestInstance(Algorithm.A, 3200, 0.5f),
            //new TestInstance(Algorithm.A, 3200, 0.7f),
            //new TestInstance(Algorithm.A, 3200, 0.9f),

            /*
            new TestInstance(Algorithm.B, 800, 20f),
            new TestInstance(Algorithm.B, 800, 30f),
            new TestInstance(Algorithm.B, 800, 40f),

            new TestInstance(Algorithm.B, 1600, 20f),
            new TestInstance(Algorithm.B, 1600, 30f),
            new TestInstance(Algorithm.B, 1600, 40f),

            new TestInstance(Algorithm.B, 2400, 20f),
            new TestInstance(Algorithm.B, 2400, 30f),
            new TestInstance(Algorithm.B, 2400, 40f),

            new TestInstance(Algorithm.B, 3200, 20f),
            new TestInstance(Algorithm.B, 3200, 30f),
            new TestInstance(Algorithm.B, 3200, 40f),


            new TestInstance(Algorithm.C, 800),
            new TestInstance(Algorithm.C, 1600),
            new TestInstance(Algorithm.C, 2400),
            new TestInstance(Algorithm.C, 3200)
            */
            };

        StartNextTest();
    }

    private void StartNextTest()
    {
        testIndex++;
        TestInstance test = tests[testIndex];

        Log("Starting test " + test.algorithm + " " + test.nrOfRays + " " + test.param);

        NUMBER_OF_RAYS = test.nrOfRays;
        algorithm = test.algorithm;

        switch (test.algorithm)
        {
            case Algorithm.DT:
                DISTANCE_TRESH = test.param;
                break;
            case Algorithm.AT:
                ANGLE_THRESH = (int) test.param;
                break;
        }

        LOG_PATH = @"c:\Users\gusaa79\TEST\" + algorithm + "_" + NUMBER_OF_RAYS;
        switch (algorithm)
        {
            case Algorithm.DT:
                LOG_PATH += "_" + DISTANCE_TRESH.ToString().Replace(".", "");
                break;
            case Algorithm.AT:
                LOG_PATH += "_" + ANGLE_THRESH;
                break;
        }
        LOG_PATH += ".json";
        windowSize = new Vector2(Screen.width, Screen.height);
        if (algorithm == Algorithm.GOLD) grid = GenerateGoldGrid();
        else grid = GenerateGrid();

        File.Delete(LOG_PATH);
        rotation = 0;
        transform.rotation = Quaternion.identity;
    }

	private void Update()
    {
        deltaTime += (Time.unscaledDeltaTime - deltaTime) * 0.1f;
        if (windowSize.x != Screen.width || windowSize.y != Screen.height)
        {
            if (algorithm == Algorithm.GOLD) grid = GenerateGoldGrid();
        	else grid = GenerateGrid();
        }
    }

	void DrawRay (Vector3 origin, Vector3 direction, Color color)
	{
		if (debug)
			Debug.DrawRay (origin, direction, color);
	}

	void DrawLine (Vector3 start, Vector3 end, Color color)
	{
		if (debug)
			Debug.DrawLine (start, end, color);
	}

	void Log (object message)
	{
		if (debug)
			Debug.Log (message);
	}

	void FixedUpdate ()
	{
        
		UpdatePlayerLookingRay ();
		DrawRay (playerLookingRay.origin, playerLookingRay.direction * 50, Color.red);

		CastRays (grid);
		CastBackgroundRays (backgroundGrid);

		buffer.Add (new JSONEntry (visibleObjects, backgroundObjects, 1.0f / deltaTime, transform));

		visibleObjects.Clear ();
		backgroundObjects.Clear ();
		if (buffer.Count == 40) {
			SaveData ();
		}
	}

	// Save the data to log
	void SaveData ()
	{
		string bufferString = "";

		List<JSONEntry> copy = new List<JSONEntry> ();
		copy.AddRange (buffer);
		buffer.Clear ();

		// copy.Sort((x, y) => x.timestamp > y.timestamp));

		foreach (JSONEntry entry in copy) {
			bufferString += entry.toJson () + "\n";
		}

		File.AppendAllText (LOG_PATH, bufferString);
	}

	void UpdatePlayerLookingRay ()
	{
		Vector3 pointOnScreen = new Vector3 (Screen.width / 2F, Screen.height / 2F, 0);
		Ray ray = Camera.main.ScreenPointToRay (pointOnScreen);
		playerLookingRay = ray;
	}


	void RotateRay (ref Ray ray, float degrees)
	{

		float rad = degrees * Mathf.PI / 180;
		ray.direction = new Vector3 (
			Mathf.Cos (rad) * ray.direction.x + Mathf.Sin (rad) * ray.direction.z,
			ray.direction.y,
			-Mathf.Sin (rad) * ray.direction.x + Mathf.Cos (rad) * ray.direction.z);
	}

   

	// Calculate a Ray's angular offset from the player's camera's center
	float CalculateAngleOffset (Ray ray)
	{
		return Vector3.Angle (playerLookingRay.direction, ray.direction);
	}

	// Checks if a renderer is in our field of view (regardless of if it is behind other objects)
	bool IsInView (Renderer  renderer)
	{
		Plane[] planes = GeometryUtility.CalculateFrustumPlanes (Camera.main);
		return GeometryUtility.TestPlanesAABB (planes, renderer.bounds);
	}

	bool IsValidObject (GameObject gameObjectQueried, GameObject gameObjectHit)
	{
		if (gameObjectHit.name == null)
			return false;
		else if (gameObjectQueried.name == "Terrain" || gameObjectQueried.name == "Plane")
			return false;
		else if (gameObjectHit.name == "FPSController" || gameObjectHit.name == "Terrain" || gameObjectHit.name == "Plane")
			return true;
		return gameObjectQueried.name == gameObjectHit.name;
	}

	List<Vector3> PointsOfInterest (Renderer renderer)
	{
		Vector3 minPoint = renderer.bounds.min;
		Vector3 maxPoint = renderer.bounds.max;

		List<Vector3> corners = new List<Vector3> {
			minPoint,
			maxPoint,
			new Vector3 (minPoint.x, minPoint.y, maxPoint.z),
			new Vector3 (minPoint.x, maxPoint.y, minPoint.z),
			new Vector3 (maxPoint.x, minPoint.y, minPoint.z),
			new Vector3 (minPoint.x, maxPoint.y, maxPoint.z),
			new Vector3 (maxPoint.x, minPoint.y, maxPoint.z),
			new Vector3 (maxPoint.x, maxPoint.y, minPoint.z),
			renderer.bounds.center
		};

		return corners;
	}

	void CastBackgroundRays (List<Vector3> backgroundGrid)
	{
		Camera camera = Camera.main;
		int raysPerDimension = (int)Math.Floor (Math.Sqrt (NUMBER_OF_RAYS));
		float vFOVrad = camera.fieldOfView * Mathf.Deg2Rad;
		float cameraHeightAt1 = Mathf.Tan (vFOVrad * .5f); 
		float hFOV = Mathf.Atan (cameraHeightAt1 * camera.aspect) * 2f * Mathf.Rad2Deg;
		float angularOffset = (360 - hFOV) / raysPerDimension;

		float playerRadius = -1;
		List<Ray> raySector = new List<Ray> ();
		foreach (Vector3 screenPoint in backgroundGrid) {
			Ray ray = camera.ScreenPointToRay (screenPoint);

			if (playerRadius < 0) {
				playerRadius = Vector3.Distance (ray.origin, transform.position);
			}
			raySector.Add (ray);
		}


		for (int sector = 0; sector <= raysPerDimension; ++sector) {
			for (int rayIndex = 0; rayIndex < raySector.Count; ++rayIndex) {
				Ray ray = raySector [rayIndex];
				if (rayIndex % 2 == 0)
					RotateRay (ref ray, angularOffset * sector + hFOV / 2);
				else if (sector == raysPerDimension)
					continue;
				else
					RotateRay (ref ray, angularOffset * sector + hFOV / 2 + angularOffset / 2);

				ray.origin = transform.position + ray.direction.normalized * playerRadius / 2;
				ray.origin = new Vector3 (ray.origin.x, camera.transform.position.y, ray.origin.z);

				RaycastHit hit;
				DrawRay (ray.origin, ray.direction, Color.green);
				if (Physics.Raycast (ray, out hit)) {
					float angleOffset = CalculateAngleOffset (ray);
					GameObject _object = hit.transform.gameObject;
					ObjectData objectData;
					if (!visibleObjects.ContainsKey (_object.GetInstanceID ())) {
						if (backgroundObjects.TryGetValue (_object.GetInstanceID (), out objectData)) {
							if (angleOffset < objectData.angle)
								objectData.angle = angleOffset;
							if (hit.distance < objectData.distance)
								objectData.distance = hit.distance;
						} else {
							backgroundObjects.Add (_object.GetInstanceID (), new ObjectData (_object) {
								angle = angleOffset,
								distance = hit.distance
							});
						}
					}
				}
			}	
		}
	}

	void CastRays(List<Vector3> grid)
    {
        Camera camera = Camera.main;
        List<double> hitDistances = new List<double>();
        Dictionary<Vector3, Vector3> hitMap = new Dictionary<Vector3, Vector3>();
        foreach (Vector3 point in grid)
        {
            Ray ray = camera.ScreenPointToRay(point);
            RaycastHit hit;
            if (Physics.Raycast(ray, out hit))
            { 
                if(hit.transform.gameObject.name != "Terrain") hitDistances.Add(hit.distance);
                DrawRay(ray.origin, ray.direction * 50, Color.green);
                hitMap.Add(point, hit.point);
                float angleOffset = CalculateAngleOffset(ray);
                GameObject _object = hit.transform.gameObject;
                ObjectData objectData;
                if (visibleObjects.TryGetValue(_object.GetInstanceID(), out objectData))
                {
                    if (angleOffset < objectData.angle) objectData.angle = angleOffset;
                    if (hit.distance < objectData.distance) objectData.distance = hit.distance;
                }
                else
                { 
                    visibleObjects.Add(_object.GetInstanceID(), new ObjectData(_object) { angle = angleOffset, distance = hit.distance, volume=CalculateVolume(gameObject)});
                }
             } else
            {
                DrawRay(ray.origin, ray.direction*50, Color.magenta);
            }
        }


        if (hitDistances.Count == 0) return;
        double distanceThreshold = GetDistanceThreshold(hitDistances);
        switch (algorithm)
        {
            case Algorithm.DT:
                AlgorithmDT(distanceThreshold);
                break;
            case Algorithm.AT:
                AlgorithmAT();
                break;
            case Algorithm.DS:
                AlgorithmDS(hitMap, distanceThreshold);
                break;
            default:
                break;
        }
    }

	List<Vector3> GenerateGrid ()
	{
		int raysPerDimension = (int)Math.Floor (Math.Sqrt (NUMBER_OF_RAYS));

		List<Vector3> grid = new List<Vector3> ();
        
		float scaleX = Screen.width / raysPerDimension;
		float scaleY = Screen.height / raysPerDimension;
		float spacingX = scaleX / 2;
		float spacingY = scaleY / 2;

		for (int x = 0; x < raysPerDimension; x++) {
			for (int y = 0; y < raysPerDimension; y++) {
				Vector3 point = new Vector3 (scaleX * x + spacingX, scaleY * y + spacingY, 0);

				AdjustVectorTowardCenter (ref point);
				grid.Add (point);
			}
		}
		backgroundGrid = new List<Vector3> ();
		for (int y = 0; y < raysPerDimension; ++y) {
			Vector3 point = new Vector3 (Screen.width / 2, scaleY * y + spacingY, 0);
			backgroundGrid.Add (point);
		}

		return grid;
	}


    List<Vector3> GenerateGoldGrid()
    {

        List<Vector3> grid = new List<Vector3>();

        float scaleY = (float) Screen.height / (float) NUMBER_OF_RAYS;
        float spacingY = scaleY / 2;
        Log("Height: " + Screen.height);
        Log("Sacel y: " + scaleY);
        for (int y = 0; y < NUMBER_OF_RAYS; y++)
        {
            Vector3 point = new Vector3(Screen.width/2, scaleY * y + spacingY, 0);
            grid.Add(point);
        }
        

        return grid;
    }

	void AdjustVectorTowardCenter (ref Vector3 vector)
	{
		Vector3 center = new Vector3 (Screen.width / 2, Screen.height / 2);
		Vector3 difference = center - vector;

		vector = vector + difference * GetGaussianValue (vector);     
	}

	float GetGaussianValue (Vector3 point)
	{
		float scale = 0.9f;
		float centerX = Mathf.Pow (point.x - Screen.width / 2, 2) / (2 * Mathf.Pow (Screen.width / 8, 2));
		float centerY = Mathf.Pow (point.y - Screen.height / 2, 2) / (2 * Mathf.Pow (Screen.height / 8, 2));
		return scale * Mathf.Exp (-(centerX + centerY));
	}

	 float CalculateVolume(GameObject gameObject)
    {
        if (gameObject.GetComponent<Renderer>() == null) return 0;
        Bounds bounds = gameObject.GetComponent<Renderer>().bounds;
        return bounds.size.x * bounds.size.y * bounds.size.z;
    }


	double GetDistanceThreshold (List<double> distances)
	{
		int stopIndex = (int)Math.Floor (distances.Count * DISTANCE_TRESH) - 1;
		distances.Sort ();
		return distances [stopIndex];
	}

	bool GameObjectIsInteresting (double threshold, GameObject gameObject)
	{
		if (visibleObjects.ContainsKey (gameObject.GetInstanceID ()) || gameObject.GetComponent<Renderer> () == null)
			return false;

		float distance = Vector3.Distance (gameObject.GetComponent<Renderer> ().bounds.center, Camera.main.transform.position);
		bool returnVal = distance < threshold && IsInView (gameObject.GetComponent<Renderer> ());
		return returnVal;
	}

	void AlgorithmDT(double threshold)
    {
        float vFOVrad = camera.fieldOfView * Mathf.Deg2Rad;
		float cameraHeightAt1 = Mathf.Tan (vFOVrad * .5f); 
		float hFOV = Mathf.Atan (cameraHeightAt1 * camera.aspect) * 2f * Mathf.Rad2Deg;
		
        foreach(GameObject gameObject in FindObjectsOfType<GameObject>())
        {
            //Skip if already tagged by point-cloud or the gameObject has no renderer-component (invisible)
            if(GameObjectIsInteresting(threshold, gameObject))
            {
                Vector3 origin = playerLookingRay.origin;
                Vector3 direction = gameObject.GetComponent<Renderer>().bounds.center - origin;

                RaycastHit hit;
                Ray ray = new Ray(origin, direction);
                float offset = CalculateAngleOffset(ray);
                
                //Fix for preventing rays outside camera view
                if (offset > hFOV/2) continue;
                
                if (Physics.Raycast(ray, out hit) && hit.transform.gameObject.name == gameObject.name)
                {
                    DrawRay(origin, direction, Color.cyan);
                    visibleObjects.Add(gameObject.GetInstanceID(), new ObjectData(gameObject) { angle = offset, distance = hit.distance, volume = CalculateVolume(gameObject) });
                }
            }
        }
    }


	void AlgorithmAT()
    {
        foreach(GameObject gameObject in FindObjectsOfType<GameObject>())
        {
            bool shouldCheck = true;
            Vector3 origin;
            Vector3 direction;
            Ray ray;

            if (visibleObjects.ContainsKey(gameObject.GetInstanceID()) || gameObject.GetComponent<Renderer>() == null || !IsInView(gameObject.GetComponent<Renderer>())) continue;

            foreach(Vector3 point in PointsOfInterest(gameObject.GetComponent<Renderer>()))
            {
                origin = playerLookingRay.origin;
                direction = point - origin;

                ray = new Ray(origin, direction);
                float angle = CalculateAngleOffset(ray);

                if (angle > ANGLE_THRESH)
                {
                    shouldCheck = false;
                    break;
                }
            }

            if (!shouldCheck) continue;
            origin = playerLookingRay.origin;
            direction = gameObject.GetComponent<Renderer>().bounds.center - origin;

            RaycastHit hit;
            ray = new Ray(origin, direction);

            if (Physics.Raycast(ray, out hit) && hit.transform.gameObject.name == gameObject.name)
            {
                DrawRay(ray.origin, ray.direction * 50, Color.green);

                visibleObjects.Add(gameObject.GetInstanceID(), new ObjectData(gameObject) {distance = hit.distance, angle = CalculateAngleOffset(ray), volume = CalculateVolume(gameObject)});
            }
            else
            {
                DrawRay(ray.origin, ray.direction * 50, Color.cyan);

            }
        }
    }

	void AlgorithmDS (Dictionary<Vector3, Vector3> hitMap, double distanceThreshold)
	{

		List<Vector3> pointsToAdd = new List<Vector3> () {
			new Vector3 (0, 0),
			new Vector3 (0, Screen.height),
			new Vector3 (Screen.width, 0),
			new Vector3 (Screen.width, Screen.height)
		};

		foreach (Vector3 point in pointsToAdd) {
			Ray ray = Camera.main.ScreenPointToRay (point);
			hitMap.Add (point, ray.GetPoint ((float)distanceThreshold));
		}

		Polygon polygon = new Polygon ();
		foreach (Vector3 point in hitMap.Keys) { 
			polygon.Add (new Vertex (point.x, point.y));
		}
		TriangleNet.Meshing.ConstraintOptions options = new TriangleNet.Meshing.ConstraintOptions () { ConformingDelaunay = false };
		mesh = (TriangleNet.Mesh)polygon.Triangulate (options);
        
		foreach (Triangle triangle in mesh.triangles) {
			Vector3 p0;
			hitMap.TryGetValue (new Vector3 ((float)triangle.GetVertex (0).x, (float)triangle.GetVertex (0).y), out p0);

			Vector3 p1;
			hitMap.TryGetValue (new Vector3 ((float)triangle.GetVertex (1).x, (float)triangle.GetVertex (1).y), out p1);

			Vector3 p2;
			hitMap.TryGetValue (new Vector3 ((float)triangle.GetVertex (2).x, (float)triangle.GetVertex (2).y), out p2);
			
            DrawLine(p0, p1, Color.white);
            DrawLine(p1, p2, Color.white);
            DrawLine(p2, p0, Color.white);
        }

        foreach (GameObject gameObject in FindObjectsOfType<GameObject>())
        {
            if (gameObject.GetComponent<Renderer>() == null || Vector3.Distance(playerLookingRay.origin, gameObject.GetComponent<Renderer>().bounds.center) > distanceThreshold) continue;

            if (visibleObjects.ContainsKey(gameObject.GetInstanceID()) || !IsInView(gameObject.GetComponent<Renderer>())) continue;

            Vector3 center = gameObject.GetComponent<Renderer>().bounds.center;
            Vector3 screencoordinates = Camera.main.WorldToScreenPoint(center);

            foreach (Triangle triangle in mesh.triangles)
            {
                Rectangle bounds = triangle.Bounds();
                if (!bounds.Contains(new Point(screencoordinates.x, screencoordinates.y))) continue;

                if (!triangle.Contains(new Point(screencoordinates.x, screencoordinates.y))) continue;

                Vector3 p0;
                hitMap.TryGetValue(new Vector3((float)triangle.GetVertex(0).x, (float)triangle.GetVertex(0).y), out p0);

                Vector3 p1;
                hitMap.TryGetValue(new Vector3((float)triangle.GetVertex(1).x, (float)triangle.GetVertex(1).y), out p1);

                Vector3 p2;
                hitMap.TryGetValue(new Vector3((float)triangle.GetVertex(2).x, (float)triangle.GetVertex(2).y), out p2);

                Plane trianglePlane = new Plane(p0, p1, p2);

                Ray ray = new Ray(playerLookingRay.origin,(center-playerLookingRay.origin));
                float rayDistance = 0.0f;

                trianglePlane.Raycast(ray, out rayDistance);

                Vector3 hitPoint = ray.GetPoint(rayDistance);
                rayDistance = Vector3.Distance(hitPoint, Camera.main.transform.position);
                if(screencoordinates.z < rayDistance)
                {
                    DrawRay(ray.origin,ray.direction*50, Color.blue);
                    RaycastHit hit;
                    if (Physics.Raycast(ray, out hit) && hit.transform.gameObject.name == gameObject.name)
                    {
                        DrawLine(p0, p1, Color.green);
                        DrawLine(p1, p2, Color.green);
                        DrawLine(p2, p0, Color.green);
                        visibleObjects.Add(gameObject.GetInstanceID(), new ObjectData(gameObject) { distance = screencoordinates.z, angle = CalculateAngleOffset(ray), volume = CalculateVolume(gameObject) });
                        //Likely visible. Add object to visible objects
                    } else
                    {
                        DrawLine(p0, p1, Color.yellow);
                        DrawLine(p1, p2, Color.yellow);
                        DrawLine(p2, p0, Color.yellow);
                    }
                }
                else
                {
                    DrawLine(p0, p1, Color.red);
                    DrawLine(p1, p2, Color.red);
                    DrawLine(p2, p0, Color.red);
                }
                break;
            }
        }
    }
}
