﻿using System;
using System.Collections.Generic;
using System.Globalization;
using UnityEngine;

[Serializable]
public class JSONEntry {
    public string timestamp;
    public float fps;
    public Vector3 position;
    public Vector3 eulerAngles;
    public List<ObjectData> visibleObjects = new List<ObjectData>();
	public List<ObjectData> peripheralObjects = new List<ObjectData>();

	public JSONEntry(
		Dictionary<int, ObjectData> _visibleObjects,
		Dictionary<int, ObjectData> _peripheralObjects,
		float _fps,
		Transform transform)
    {
        visibleObjects.AddRange(_visibleObjects.Values);
		peripheralObjects.AddRange(_peripheralObjects.Values);
        timestamp = DateTime.UtcNow.ToString("yyyy-MM-dd HH:mm:ss.fff", CultureInfo.InvariantCulture);

        fps = _fps;
        position = transform.position;
        eulerAngles = transform.rotation.eulerAngles;
    }

    public string toJson()
    {
        return JsonUtility.ToJson(this);
    }

    [Serializable]
    public class YawPitchRoll
    {
        public float yaw;
        public float pitch;
        public float roll;

        public YawPitchRoll(Transform transform)
        {
            yaw = transform.rotation.eulerAngles.y;
            pitch = getPitch(transform);
            roll = getRoll(transform);
        }

        float getYaw(Transform t)
        {
            return t.localEulerAngles.z;
        }

        float getPitch(Transform originalTransform)
        {
            GameObject tempGO = new GameObject();
            Transform t = tempGO.transform;
            t.localRotation = originalTransform.localRotation;

            t.Rotate(0, 0, t.localEulerAngles.z * -1);
            t.Rotate(t.localEulerAngles.x * -1, 0, 0);

            GameObject.Destroy(tempGO);
            return t.localEulerAngles.y;
        }

        float getRoll(Transform originalTransform)
        {
            GameObject tempGO = new GameObject();
            Transform t = tempGO.transform;
            t.localRotation = originalTransform.localRotation;

            t.Rotate(0, 0, t.localEulerAngles.z * -1);

            GameObject.Destroy(tempGO);
            return t.localEulerAngles.x;
        }
    }
}
